package ByteCode;
import java.util.*;
import ClassFile.*;
import DataFlow.*;
import Misc.*;

public abstract class Instruction implements ByteCodeConst, ClassFileConst, DataFlowConst
{
  protected ByteCode bc;
  public int offset;
  public int opcode;
  public int length;

  //
  // Loading
  //
  public Instruction(ByteCode owner, int ofs) {
    bc = owner;
    offset = ofs;
    opcode = util.b2i(bc.code[ofs]);
  }

  //
  // Saving
  //
  public void Adjust(int ofs) {			// to be overrided
    offset = ofs;
  }

  public void Save() {				// to be overrided
    WriteByteU(offset, opcode);
  }

  //
  // DataFlow
  //
  public int pc;				// opcode# from start
  public boolean fSaveState;			// save state or not
  public State state;				// state

  protected boolean SaveState(State stat) throws DataFlowException {
    if ( state != null )
      if ( stat.Merge(state) )
	return true;
    if ( fSaveState )
	state = new State(stat);
    return false;
  }

  protected static String strOpcode = "unexpected opcode";
  protected static String strType   = "type mismatch";

  public abstract NextState Analyze(State stat)	// to be implemented
    throws DataFlowException;

  //
  // CodeModify
  //
  public Instruction(int op) {			// to be overrided
    opcode = op;
    offset = -1;
  }

  protected void SetOwner(ByteCode owner) {
    bc = owner;
  }

  //
  // utilities
  //
  protected boolean InShortRange(Label label) {
    int idiff = label.offset - offset;
    short sdiff = (short)idiff;
    return (idiff == sdiff);
  }

  protected short RelShort(Label label) {
    return (short)(label.offset - offset);
  }

  protected int RelInt(Label label) {
    return label.offset - offset;
  }

  protected int ReadByteS(int ofs) {
    return bc.code[ofs];
  }

  protected int ReadByteU(int ofs) {
    return util.b2i(bc.code[ofs]);
  }

  protected int ReadShort(int ofs) {
    return (short)((util.b2i(bc.code[ofs]) << 8) | util.b2i(bc.code[ofs+1]));
  }

  protected int ReadInt(int ofs) {
    return (util.b2i(bc.code[ofs]) << 24) | (util.b2i(bc.code[ofs+1]) << 16)
	   | (util.b2i(bc.code[ofs+2]) << 8) | util.b2i(bc.code[ofs+3]);
  }

  protected void WriteByteS(int ofs, int value) {
    bc.code[ofs] = (byte)value;
  }

  protected void WriteByteU(int ofs, int value) {
    bc.code[ofs] = (byte)value;
  }

  protected void WriteShort(int ofs, int value) {
    bc.code[ofs]	= (byte)(value >>> 8);
    bc.code[ofs+1]	= (byte)value;
  }

  protected void WriteInt(int ofs, int value) {
    bc.code[ofs]	= (byte)(value >>> 24);
    bc.code[ofs+1]	= (byte)(value >>> 16);
    bc.code[ofs+2]	= (byte)(value >>> 8);
    bc.code[ofs+3]	= (byte)value;
  }

  //
  // opcode classification
  //
  static int[] optype = new int[256];

  static public Instruction AutoLoad(ByteCode bc, int ofs) throws ByteCodeException {
    switch ( optype[util.b2i(bc.code[ofs])] ) {
    case OP_NOP:	  return new opNOP(bc, ofs);
    case OP_CONST:	  return new opCONST(bc, ofs);
    case OP_PUSH:	  return new opPUSH(bc, ofs);
    case OP_LDC:	  return new opLDC(bc, ofs);
    case OP_LOAD:	  return new opLOAD(bc, ofs);
    case OP_LOAD2:	  return new opLOAD2(bc, ofs);
    case OP_LOADA:	  return new opLOADA(bc, ofs);
    case OP_STORE:	  return new opSTORE(bc, ofs);
    case OP_STORE2:	  return new opSTORE2(bc, ofs);
    case OP_STOREA:	  return new opSTOREA(bc, ofs);
    case OP_POP:	  return new opPOP(bc, ofs);
    case OP_DUP:	  return new opDUP(bc, ofs);
    case OP_SWAP:	  return new opSWAP(bc, ofs);
    case OP_ARITH:	  return new opARITH(bc, ofs);
    case OP_IINC:	  return new opIINC(bc, ofs);
    case OP_CONV:	  return new opCONV(bc, ofs);
    case OP_COMP:	  return new opCOMP(bc, ofs);
    case OP_COND:	  return new opCOND(bc, ofs);
    case OP_GOTO:	  return new opGOTO(bc, ofs);
    case OP_JSR:	  return new opJSR(bc, ofs);
    case OP_RET:	  return new opRET(bc, ofs);
    case OP_TABLE:	  return new opTABLE(bc, ofs);
    case OP_LOOKUP:	  return new opLOOKUP(bc, ofs);
    case OP_RETURN:	  return new opRETURN(bc, ofs);
    case OP_FIELD:	  return new opFIELD(bc, ofs);
    case OP_INVOKE:	  return new opINVOKE(bc, ofs);
    case OP_NEW:	  return new opNEW(bc, ofs);
    case OP_NEWARRAY:	  return new opNEWARRAY(bc, ofs);
    case OP_ARRAYLENGTH:  return new opARRAYLENGTH(bc, ofs);
    case OP_ATHROW:	  return new opATHROW(bc, ofs);
    case OP_CHECKCAST:    return new opCHECKCAST(bc, ofs);
    case OP_INSTANCEOF:   return new opINSTANCEOF(bc, ofs);
    case OP_MONITOR:      return new opMONITOR(bc, ofs);
    case OP_WIDE:	  return new opWIDE(bc, ofs);
    default: 		  break;
    }
    throw new ByteCodeException("Instruction.AutoLoad", "invalid opcode: " + util.b2i(bc.code[ofs]));
  }

  static {
    optype[NOP]			= OP_NOP;
    optype[ACONST_NULL]		= OP_CONST;
    optype[ICONST_M1]		= OP_CONST;
    optype[ICONST_0]		= OP_CONST;
    optype[ICONST_1]		= OP_CONST;
    optype[ICONST_2]		= OP_CONST;
    optype[ICONST_3]		= OP_CONST;
    optype[ICONST_4]		= OP_CONST;
    optype[ICONST_5]		= OP_CONST;
    optype[LCONST_0]		= OP_CONST;
    optype[LCONST_1]		= OP_CONST;
    optype[FCONST_0]		= OP_CONST;
    optype[FCONST_1]		= OP_CONST;
    optype[FCONST_2]		= OP_CONST;
    optype[DCONST_0]		= OP_CONST;
    optype[DCONST_1]		= OP_CONST;
    optype[BIPUSH]		= OP_PUSH;
    optype[SIPUSH]		= OP_PUSH;
    optype[LDC]			= OP_LDC;
    optype[LDC_W]		= OP_LDC;
    optype[LDC2_W]		= OP_LDC;
    optype[ILOAD]		= OP_LOAD;
    optype[LLOAD]		= OP_LOAD;
    optype[FLOAD]		= OP_LOAD;
    optype[DLOAD]		= OP_LOAD;
    optype[ALOAD]		= OP_LOAD;
    optype[ILOAD_0]		= OP_LOAD2;
    optype[ILOAD_1]		= OP_LOAD2;
    optype[ILOAD_2]		= OP_LOAD2;
    optype[ILOAD_3]		= OP_LOAD2;
    optype[LLOAD_0]		= OP_LOAD2;
    optype[LLOAD_1]		= OP_LOAD2;
    optype[LLOAD_2]		= OP_LOAD2;
    optype[LLOAD_3]		= OP_LOAD2;
    optype[FLOAD_0]		= OP_LOAD2;
    optype[FLOAD_1]		= OP_LOAD2;
    optype[FLOAD_2]		= OP_LOAD2;
    optype[FLOAD_3]		= OP_LOAD2;
    optype[DLOAD_0]		= OP_LOAD2;
    optype[DLOAD_1]		= OP_LOAD2;
    optype[DLOAD_2]		= OP_LOAD2;
    optype[DLOAD_3]		= OP_LOAD2;
    optype[ALOAD_0]		= OP_LOAD2;
    optype[ALOAD_1]		= OP_LOAD2;
    optype[ALOAD_2]		= OP_LOAD2;
    optype[ALOAD_3]		= OP_LOAD2;
    optype[IALOAD]		= OP_LOADA;
    optype[LALOAD]		= OP_LOADA;
    optype[FALOAD]		= OP_LOADA;
    optype[DALOAD]		= OP_LOADA;
    optype[AALOAD]		= OP_LOADA;
    optype[BALOAD]		= OP_LOADA;
    optype[CALOAD]		= OP_LOADA;
    optype[SALOAD]		= OP_LOADA;
    optype[ISTORE]		= OP_STORE;
    optype[LSTORE]		= OP_STORE;
    optype[FSTORE]		= OP_STORE;
    optype[DSTORE]		= OP_STORE;
    optype[ASTORE]		= OP_STORE;
    optype[ISTORE_0]		= OP_STORE2;
    optype[ISTORE_1]		= OP_STORE2;
    optype[ISTORE_2]		= OP_STORE2;
    optype[ISTORE_3]		= OP_STORE2;
    optype[LSTORE_0]		= OP_STORE2;
    optype[LSTORE_1]		= OP_STORE2;
    optype[LSTORE_2]		= OP_STORE2;
    optype[LSTORE_3]		= OP_STORE2;
    optype[FSTORE_0]		= OP_STORE2;
    optype[FSTORE_1]		= OP_STORE2;
    optype[FSTORE_2]		= OP_STORE2;
    optype[FSTORE_3]		= OP_STORE2;
    optype[DSTORE_0]		= OP_STORE2;
    optype[DSTORE_1]		= OP_STORE2;
    optype[DSTORE_2]		= OP_STORE2;
    optype[DSTORE_3]		= OP_STORE2;
    optype[ASTORE_0]		= OP_STORE2;
    optype[ASTORE_1]		= OP_STORE2;
    optype[ASTORE_2]		= OP_STORE2;
    optype[ASTORE_3]		= OP_STORE2;
    optype[IASTORE]		= OP_STOREA;
    optype[LASTORE]		= OP_STOREA;
    optype[FASTORE]		= OP_STOREA;
    optype[DASTORE]		= OP_STOREA;
    optype[AASTORE]		= OP_STOREA;
    optype[BASTORE]		= OP_STOREA;
    optype[CASTORE]		= OP_STOREA;
    optype[SASTORE]		= OP_STOREA;
    optype[POP]			= OP_POP;
    optype[POP2]		= OP_POP;
    optype[DUP]			= OP_DUP;
    optype[DUP_X1]		= OP_DUP;
    optype[DUP_X2]		= OP_DUP;
    optype[DUP2]		= OP_DUP;
    optype[DUP2_X1]		= OP_DUP;
    optype[DUP2_X2]		= OP_DUP;
    optype[SWAP]		= OP_SWAP;
    optype[IADD]		= OP_ARITH;
    optype[LADD]		= OP_ARITH;
    optype[FADD]		= OP_ARITH;
    optype[DADD]		= OP_ARITH;
    optype[ISUB]		= OP_ARITH;
    optype[LSUB]		= OP_ARITH;
    optype[FSUB]		= OP_ARITH;
    optype[DSUB]		= OP_ARITH;
    optype[IMUL]		= OP_ARITH;
    optype[LMUL]		= OP_ARITH;
    optype[FMUL]		= OP_ARITH;
    optype[DMUL]		= OP_ARITH;
    optype[IDIV]		= OP_ARITH;
    optype[LDIV]		= OP_ARITH;
    optype[FDIV]		= OP_ARITH;
    optype[DDIV]		= OP_ARITH;
    optype[IREM]		= OP_ARITH;
    optype[LREM]		= OP_ARITH;
    optype[FREM]		= OP_ARITH;
    optype[DREM]		= OP_ARITH;
    optype[INEG]		= OP_ARITH;
    optype[LNEG]		= OP_ARITH;
    optype[FNEG]		= OP_ARITH;
    optype[DNEG]		= OP_ARITH;
    optype[ISHL]		= OP_ARITH;
    optype[LSHL]		= OP_ARITH;
    optype[ISHR]		= OP_ARITH;
    optype[LSHR]		= OP_ARITH;
    optype[IUSHR]		= OP_ARITH;
    optype[LUSHR]		= OP_ARITH;
    optype[IAND]		= OP_ARITH;
    optype[LAND]		= OP_ARITH;
    optype[IOR]			= OP_ARITH;
    optype[LOR]			= OP_ARITH;
    optype[IXOR]		= OP_ARITH;
    optype[LXOR]		= OP_ARITH;
    optype[IINC]		= OP_IINC;
    optype[I2L]			= OP_CONV;
    optype[I2F]			= OP_CONV;
    optype[I2D]			= OP_CONV;
    optype[L2I]			= OP_CONV;
    optype[L2F]			= OP_CONV;
    optype[L2D]			= OP_CONV;
    optype[F2I]			= OP_CONV;
    optype[F2L]			= OP_CONV;
    optype[F2D]			= OP_CONV;
    optype[D2I]			= OP_CONV;
    optype[D2L]			= OP_CONV;
    optype[D2F]			= OP_CONV;
    optype[I2B]			= OP_CONV;
    optype[I2C]			= OP_CONV;
    optype[I2S]			= OP_CONV;
    optype[LCMP]		= OP_COMP;
    optype[FCMPL]		= OP_COMP;
    optype[FCMPG]		= OP_COMP;
    optype[DCMPL]		= OP_COMP;
    optype[DCMPG]		= OP_COMP;
    optype[IFEQ]		= OP_COND;
    optype[IFNE]		= OP_COND;
    optype[IFLT]		= OP_COND;
    optype[IFGE]		= OP_COND;
    optype[IFGT]		= OP_COND;
    optype[IFLE]		= OP_COND;
    optype[IF_ICMPEQ]		= OP_COND;
    optype[IF_ICMPNE]		= OP_COND;
    optype[IF_ICMPLT]		= OP_COND;
    optype[IF_ICMPGE]		= OP_COND;
    optype[IF_ICMPGT]		= OP_COND;
    optype[IF_ICMPLE]		= OP_COND;
    optype[IF_ACMPEQ]		= OP_COND;
    optype[IF_ACMPNE]		= OP_COND;
    optype[GOTO]		= OP_GOTO;
    optype[JSR]			= OP_JSR;
    optype[RET]			= OP_RET;
    optype[TABLESWITCH]		= OP_TABLE;
    optype[LOOKUPSWITCH]	= OP_LOOKUP;
    optype[IRETURN]		= OP_RETURN;
    optype[LRETURN]		= OP_RETURN;
    optype[FRETURN]		= OP_RETURN;
    optype[DRETURN]		= OP_RETURN;
    optype[ARETURN]		= OP_RETURN;
    optype[RETURN]		= OP_RETURN;
    optype[GETSTATIC]		= OP_FIELD;
    optype[PUTSTATIC]		= OP_FIELD;
    optype[GETFIELD]		= OP_FIELD;
    optype[PUTFIELD]		= OP_FIELD;
    optype[INVOKEVIRTUAL]	= OP_INVOKE;
    optype[INVOKESPECIAL]	= OP_INVOKE;
    optype[INVOKESTATIC]	= OP_INVOKE;
    optype[INVOKEINTERFACE]	= OP_INVOKE;
    optype[186]			= OP_INVALID;
    optype[NEW]			= OP_NEW;
    optype[NEWARRAY]		= OP_NEWARRAY;
    optype[ANEWARRAY]		= OP_NEWARRAY;
    optype[ARRAYLENGTH]		= OP_ARRAYLENGTH;
    optype[ATHROW]		= OP_ATHROW;
    optype[CHECKCAST]		= OP_CHECKCAST;
    optype[INSTANCEOF]		= OP_INSTANCEOF;
    optype[MONITORENTER]	= OP_MONITOR;
    optype[MONITOREXIT]		= OP_MONITOR;
    optype[WIDE]		= OP_WIDE;
    optype[MULTIANEWARRAY]	= OP_NEWARRAY;
    optype[IFNULL]		= OP_COND;
    optype[IFNONNULL]		= OP_COND;
    optype[GOTO_W]		= OP_GOTO;
    optype[JSR_W]		= OP_JSR;
    optype[202]			= OP_INVALID;
    optype[203]			= OP_INVALID;
    optype[204]			= OP_INVALID;
    optype[205]			= OP_INVALID;
    optype[206]			= OP_INVALID;
    optype[207]			= OP_INVALID;
    optype[208]			= OP_INVALID;
    optype[209]			= OP_INVALID;
    optype[210]			= OP_INVALID;
    optype[211]			= OP_INVALID;
    optype[212]			= OP_INVALID;
    optype[213]			= OP_INVALID;
    optype[214]			= OP_INVALID;
    optype[215]			= OP_INVALID;
    optype[216]			= OP_INVALID;
    optype[217]			= OP_INVALID;
    optype[218]			= OP_INVALID;
    optype[219]			= OP_INVALID;
    optype[220]			= OP_INVALID;
    optype[221]			= OP_INVALID;
    optype[222]			= OP_INVALID;
    optype[223]			= OP_INVALID;
    optype[224]			= OP_INVALID;
    optype[225]			= OP_INVALID;
    optype[226]			= OP_INVALID;
    optype[227]			= OP_INVALID;
    optype[228]			= OP_INVALID;
    optype[229]			= OP_INVALID;
    optype[230]			= OP_INVALID;
    optype[231]			= OP_INVALID;
    optype[232]			= OP_INVALID;
    optype[233]			= OP_INVALID;
    optype[234]			= OP_INVALID;
    optype[235]			= OP_INVALID;
    optype[236]			= OP_INVALID;
    optype[237]			= OP_INVALID;
    optype[238]			= OP_INVALID;
    optype[239]			= OP_INVALID;
    optype[240]			= OP_INVALID;
    optype[241]			= OP_INVALID;
    optype[242]			= OP_INVALID;
    optype[243]			= OP_INVALID;
    optype[244]			= OP_INVALID;
    optype[245]			= OP_INVALID;
    optype[246]			= OP_INVALID;
    optype[247]			= OP_INVALID;
    optype[248]			= OP_INVALID;
    optype[249]			= OP_INVALID;
    optype[250]			= OP_INVALID;
    optype[251]			= OP_INVALID;
    optype[252]			= OP_INVALID;
    optype[253]			= OP_INVALID;
    optype[254]			= OP_INVALID;
    optype[255]			= OP_INVALID;
  }
}
