package DataFlow;
import java.util.*;
import ByteCode.*;
import ClassFile.*;
import CodeModify.*;

public class Var implements ByteCodeConst, ClassFileConst, DataFlowConst
{
  private static Unifier unif = new Unifier();
  private int	 type;
  private String desc;
  private Label  subr;

  public static Var vinv, vbool, vchar, vfloat, vdouble;
  public static Var vbyte, vshort, vint, vlong;
  public static Var vnull, vstring, vthrow;

  private Var(int t) {
    type = t;
  }

  private Var(String des) {
    type = T_REFER;
    desc = new String(des);
  }

  public Var(Label subroutine) {
    type = T_RETURN;
    subr = subroutine;
  }

  //
  // constructions
  //
  public static Var getVar(String desc) throws DataFlowException {
    switch ( desc.charAt(0) ) {
    case DC_BYTE:    return vbyte;
    case DC_CHAR:    return vchar;
    case DC_DOUBLE:  return vdouble;
    case DC_FLOAT:   return vfloat;
    case DC_INT:     return vint;
    case DC_LONG:    return vlong;
    case DC_SHORT:   return vshort;
    case DC_BOOLEAN: return vbool;
    case DC_VOID:    return null;
    case DC_CLASS_BEGIN:
    case DC_ARRAY:   return new Var(desc);
    default:	     throw new DataFlowException("Var.getVar", "desc: " + desc);
    }
  }

  public static Vector getVars(String desc) throws DataFlowException {
    int i = 0, len = desc.length();
    char ch;
    Vector v = new Vector();
    String s;
    while ( i < len ) {
      for (s = ""; ; i++) {
	switch ( (ch = desc.charAt(i)) ) {
	case DC_BYTE:  case DC_CHAR:	case DC_DOUBLE:
	case DC_FLOAT: case DC_INT:	case DC_LONG:
	case DC_SHORT: case DC_BOOLEAN:	case DC_VOID:
	  s += ch;
	  break;
	case DC_CLASS_BEGIN:
	  for (; desc.charAt(i) != DC_CLASS_END; i++)
	    s += desc.charAt(i);
	  s += DC_CLASS_END;
	  break;
	case DC_ARRAY:
	  s += ch;
	  continue;
	}
	break;
      }
      v.addElement(getVar(s)); i++;
    }
    return v;
  }

  public static Var getArrayVar(int atype) throws DataFlowException {
    String desc = String.valueOf(DC_ARRAY);
    switch ( atype ) {
    case T_BOOLEAN: return new Var(desc + DC_BOOLEAN);
    case T_CHAR:    return new Var(desc + DC_CHAR);
    case T_FLOAT:   return new Var(desc + DC_FLOAT);
    case T_DOUBLE:  return new Var(desc + DC_DOUBLE);
    case T_BYTE:    return new Var(desc + DC_BYTE);
    case T_SHORT:   return new Var(desc + DC_SHORT);
    case T_INT:	    return new Var(desc + DC_INT);
    case T_LONG:    return new Var(desc + DC_LONG);
    default:	    throw new DataFlowException("Var.getArrayVar(int)", "atype: " + atype);
    }
  }

  public static Var getArrayVar(Var v, int dim) throws DataFlowException {
    if ( v.type != T_REFER || dim < 1 )
      throw new DataFlowException("Var.getArrayVar(Var, int)", "for multianewarray only");
    String ch = String.valueOf(DC_ARRAY), desc = new String(ch);
    for (int i = 1; i < dim; i++)
      desc += ch;
    desc += v.desc;
    return new Var(desc);
  }

  public static Var getArrayElementVar(Var v) throws DataFlowException {
    String here = "Var.getArrayElementVar";
    if ( !v.isArray() )
      throw new DataFlowException(here, null);
    return getVar(v.desc.substring(1));
  }

  //
  // operations
  //
  public boolean equals(Var v) {
    if ( this == v ) return true;
    boolean f = (type == v.type && subr == v.subr);
    if ( desc == null && v.desc == null ) return f;
    if ( desc != null && v.desc != null )
      return (f && desc.equals(v.desc));
    return false;
  }

  public boolean isArray() {
    return (type == T_REFER && desc != null &&
	    desc.charAt(0) == DC_ARRAY);
  }

  public boolean isCategory1() {
    return (type != T_LONG && type != T_DOUBLE);
  }

  public boolean isCategory2() {
    return (type == T_LONG || type == T_DOUBLE);
  }

  public boolean isCompatible(int vt) {
    if (vt == T_RETURN)
      return (type >= T_REFER);
    return (type == vt);
  }

  public boolean isCompatible(Var v) throws DataFlowException {
    if ( type != v.type )
      return false;
    if ( type == T_RETURN )
      return (subr == v.subr);
    if ( type != T_REFER )
      return true;
    return (desc == null) || unif.SubClassOf(desc, v.desc);
  }

  public boolean isSameType(int vt) {
    return (type == vt);
  }

  public boolean isSameType(Var v) {
    return (type == v.type);
  }

  public Label getSubroutine() throws DataFlowException {
    if ( type != T_RETURN )
      throw new DataFlowException("Var.getSubroutine", null);
    return subr;
  }

  //
  // Merging
  //
  public Var Merge(Var v, boolean strict) throws DataFlowException {
    String here = "Var.Merge";
    if ( type == T_INVALID || v.type == T_INVALID )
      return vinv;
    if ( type != v.type || subr != v.subr )
      if ( !strict ) return vinv;
      else throw new DataFlowException(here, "type mismatch");
    if ( type != T_REFER ) return this;
    if ( v.desc == null )  return this;
    if ( desc == null )	   return v;
    return getVar(unif.Unify(desc, v.desc));
  }

  //
  // CodeModify
  //
  public int getVMType() throws CodeModifyException {
    switch ( type ) {
    case T_BOOLEAN: case T_CHAR: case T_BYTE: case T_SHORT: case T_INT:
      return 0;
    case T_FLOAT:
      return 2;
    case T_DOUBLE:
      return 3;
    case T_LONG:
      return 1;
    case T_REFER:
      return 4;
    case T_RETURN:
      return 5;
    }
    throw new CodeModifyException("Var.getVMType", null);
  }

  public String getClassName() throws CodeModifyException {
    String strip = (desc.charAt(0) == DC_ARRAY) ? desc :
		   desc.substring(1, desc.length()-1);
    return strip;
  }

  public Instruction store(int i) throws CodeModifyException {
    switch ( type ) {
    case T_BOOLEAN: case T_CHAR: case T_BYTE: case T_SHORT: case T_INT:
      if (i < 4) return new opSTORE2(ISTORE_0 + i);
      else	 return new opSTORE(ISTORE, i);
    case T_FLOAT:
      if (i < 4) return new opSTORE2(FSTORE_0 + i);
      else	 return new opSTORE(FSTORE, i);
    case T_DOUBLE:
      if (i < 4) return new opSTORE2(DSTORE_0 + i);
      else	 return new opSTORE(DSTORE, i);
    case T_LONG:
      if (i < 4) return new opSTORE2(LSTORE_0 + i);
      else	 return new opSTORE(LSTORE, i);
    case T_REFER: case T_RETURN:
      if (i < 4) return new opSTORE2(ASTORE_0 + i);
      else	 return new opSTORE(ASTORE, i);
    }
    throw new CodeModifyException("Var.store", null);
  }

  public Instruction load(int i) throws CodeModifyException {
    switch ( type ) {
    case T_BOOLEAN: case T_CHAR: case T_BYTE: case T_SHORT: case T_INT:
      if (i < 4) return new opLOAD2(ILOAD_0 + i);
      else	 return new opLOAD(ILOAD, i);
    case T_FLOAT:
      if (i < 4) return new opLOAD2(FLOAD_0 + i);
      else	 return new opLOAD(FLOAD, i);
    case T_DOUBLE:
      if (i < 4) return new opLOAD2(DLOAD_0 + i);
      else	 return new opLOAD(DLOAD, i);
    case T_LONG:
      if (i < 4) return new opLOAD2(LLOAD_0 + i);
      else	 return new opLOAD(LLOAD, i);
    case T_REFER: case T_RETURN:
      if (i < 4) return new opLOAD2(ALOAD_0 + i);
      else	 return new opLOAD(ALOAD, i);
    }
    throw new CodeModifyException("Var.load", null);
  }

  public Instruction astore() throws CodeModifyException {
    switch ( type ) {
    case T_BOOLEAN: case T_CHAR: case T_BYTE: case T_SHORT: case T_INT:
      return new opSTOREA(IASTORE);
    case T_FLOAT:
      return new opSTOREA(FASTORE);
    case T_DOUBLE:
      return new opSTOREA(DASTORE);
    case T_LONG:
      return new opSTOREA(LASTORE);
    case T_REFER:
      return new opSTOREA(AASTORE);
    }
    throw new CodeModifyException("Var.astore", null);
  }

  public Instruction aload() throws CodeModifyException {
    switch ( type ) {
    case T_BOOLEAN: case T_CHAR: case T_BYTE: case T_SHORT: case T_INT:
      return new opLOADA(IALOAD);
    case T_FLOAT:
      return new opLOADA(FALOAD);
    case T_DOUBLE:
      return new opLOADA(DALOAD);
    case T_LONG:
      return new opLOADA(LALOAD);
    case T_REFER:
      return new opLOADA(AALOAD);
    }
    throw new CodeModifyException("Var.aload", null);
  }

  //
  // static initializing
  //
  static {
    vinv    = new Var(T_INVALID);
    vbool   = new Var(T_INT);
    vchar   = new Var(T_INT);
    vfloat  = new Var(T_FLOAT);
    vdouble = new Var(T_DOUBLE);
    vbyte   = new Var(T_INT);
    vshort  = new Var(T_INT);
    vint    = new Var(T_INT);
    vlong   = new Var(T_LONG);
    vnull   = new Var(T_REFER);
    vstring = new Var("Ljava/lang/String;");
    vthrow  = new Var("Ljava/lang/Throwable;");
  }

  //
  // debug
  //
  public String ShowDebug() {
    switch ( type ) {
    case T_INVALID: return "*";
    case T_BOOLEAN: return String.valueOf(DC_BOOLEAN);
    case T_CHAR:    return String.valueOf(DC_CHAR);
    case T_FLOAT:   return String.valueOf(DC_FLOAT);
    case T_DOUBLE:  return String.valueOf(DC_DOUBLE);
    case T_BYTE:    return String.valueOf(DC_BYTE);
    case T_SHORT:   return String.valueOf(DC_SHORT);
    case T_INT:	    return String.valueOf(DC_INT);
    case T_LONG:    return String.valueOf(DC_LONG);
    case T_REFER:   return (desc == null) ? "nil" : desc;
    case T_RETURN:   return "R("+subr.offset+")";
    }
    return "Var.ShowDebug's BUG";
  }
}

class Unifier implements ClassFileConst
{
  public Hashtable hash;

  public Unifier() {
    hash = new Hashtable();
  }

  private String getQualifiedName(String desc) {
    String strip = (desc.charAt(0) == DC_ARRAY) ? desc :
		   desc.substring(1, desc.length()-1);
    return strip.replace('/', '.');
  }

  private String getDescriptorName(String qual) {
    String wrap = (qual.charAt(0) == DC_ARRAY) ? qual :
		  DC_CLASS_BEGIN + qual + DC_CLASS_END;
    return wrap.replace('.', '/');
  }

  private Vector getClassVector(String qual) throws ClassNotFoundException {
    Vector v = (Vector)hash.get(qual);
    if ( v == null ) {
      Class clazz = Class.forName(qual);
      v = new Vector();
      do {
	v.insertElementAt(clazz.getName(), 0);
	clazz = clazz.getSuperclass();
      } while ( clazz != null );
      hash.put(qual, v);
    }
    return v;
  }

  public boolean SubClassOf(String desc1, String desc2) throws DataFlowException {
    try {
      String q1 = getQualifiedName(desc1);
      String q2 = getQualifiedName(desc2);
      Vector v1 = getClassVector(q1);
      Vector v2 = getClassVector(q2);
      if ( v1.size() < v2.size() )
	return false;
      return v1.elementAt(v2.size()-1).equals(q2);
    } catch (ClassNotFoundException e) {
      throw new DataFlowException("Unifier.SubClassOf", e.toString());
    }
  }

  public String Unify(String desc1, String desc2) throws DataFlowException {
    try {
      String q1 = getQualifiedName(desc1);
      String q2 = getQualifiedName(desc2);
      Vector v1 = getClassVector(q1);
      Vector v2 = getClassVector(q2);
      int i, sz = Math.min(v1.size(), v2.size());
      for (i = 0; i < sz; i++)
	if ( !v1.elementAt(i).equals(v2.elementAt(i)) )
	  break;
      if ( i > 0 ) i--;
      return getDescriptorName((String)v1.elementAt(i));
    } catch (ClassNotFoundException e) {
      throw new DataFlowException("Unifier.Unify", e.toString());
    }
  }
}
